@@ -49,10 +49,8 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <boost/weak_ptr.hpp>
 #include <boost/noncopyable.hpp>
 #include <boost/array.hpp>
-#include <boost/date_time/posix_time/posix_time.hpp>
 #include <boost/optional.hpp>
 #include <boost/cstdint.hpp>
-#include <boost/detail/atomic_count.hpp>
 
 #ifdef _MSC_VER
 #pragma warning(pop)
@@ -71,50 +69,71 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/piece_block_progress.hpp"
 #include "libtorrent/config.hpp"
 #include "libtorrent/session.hpp"
-
-// TODO: each time a block is 'taken over'
-// from another peer. That peer must be given
-// a chance to become not-interested.
+#include "libtorrent/bandwidth_manager.hpp"
+#include "libtorrent/policy.hpp"
+#include "libtorrent/socket_type.hpp"
+#include "libtorrent/intrusive_ptr_base.hpp"
 
 namespace libtorrent
 {
 	class torrent;
+	struct peer_plugin;
 
 	namespace detail
 	{
 		struct session_impl;
 	}
 
-	TORRENT_EXPORT void intrusive_ptr_add_ref(peer_connection const*);
-	TORRENT_EXPORT void intrusive_ptr_release(peer_connection const*);	
-
 	struct TORRENT_EXPORT protocol_error: std::runtime_error
 	{
 		protocol_error(const std::string& msg): std::runtime_error(msg) {};
 	};
 
 	class TORRENT_EXPORT peer_connection
-		: public boost::noncopyable
+		: public intrusive_ptr_base<peer_connection>
+		, public boost::noncopyable
 	{
 	friend class invariant_access;
-	friend void intrusive_ptr_add_ref(peer_connection const*);
-	friend void intrusive_ptr_release(peer_connection const*);
 	public:
 
+		enum channels
+		{
+			upload_channel,
+			download_channel,
+			num_channels
+		};
+
 		// this is the constructor where the we are the active part.
 		// The peer_conenction should handshake and verify that the
 		// other end has the correct id
 		peer_connection(
 			aux::session_impl& ses
 			, boost::weak_ptr<torrent> t
-			, boost::shared_ptr<stream_socket> s
-			, tcp::endpoint const& remote);
+			, boost::shared_ptr<socket_type> s
+			, tcp::endpoint const& remote
+			, policy::peer* peerinfo);
 
 		// with this constructor we have been contacted and we still don't
 		// know which torrent the connection belongs to
 		peer_connection(
 			aux::session_impl& ses
-			, boost::shared_ptr<stream_socket> s);
+			, boost::shared_ptr<socket_type> s
+			, policy::peer* peerinfo);
+
+		virtual ~peer_connection();
+
+		void set_peer_info(policy::peer* pi)
+		{ m_peer_info = pi; }
+
+		policy::peer* peer_info_struct() const
+		{ return m_peer_info; }
+
+		enum peer_speed_t { slow, medium, fast };
+		peer_speed_t peer_speed();
+
+#ifndef TORRENT_DISABLE_EXTENSIONS
+		void add_extension(boost::shared_ptr<peer_plugin>);
+#endif
 
 		// this function is called once the torrent associated
 		// with this peer connection has retrieved the meta-
@@ -122,10 +141,30 @@ namespace libtorrent
 		// this is called from the constructor.
 		void init();
 
+		// this is called when the metadata is retrieved
+		// and the files has been checked
+		virtual void on_metadata() {}
+
 		void set_upload_limit(int limit);
 		void set_download_limit(int limit);
 
-		virtual ~peer_connection();
+		int upload_limit() const { return m_upload_limit; }
+		int download_limit() const { return m_download_limit; }
+
+		bool prefer_whole_pieces() const
+		{ return m_prefer_whole_pieces; }
+
+		void prefer_whole_pieces(bool b)
+		{ m_prefer_whole_pieces = b; }
+
+		bool request_large_blocks() const
+		{ return m_request_large_blocks; }
+
+		void request_large_blocks(bool b)
+		{ m_request_large_blocks = b; }
+
+		void set_non_prioritized(bool b)
+		{ m_non_prioritized = b; }
 
 		// this adds an announcement in the announcement queue
 		// it will let the peer know that we have the given piece
@@ -157,6 +196,10 @@ namespace libtorrent
 		bool is_peer_interested() const { return m_peer_interested; }
 		bool has_peer_choked() const { return m_peer_choked; }
 
+		void update_interest();
+
+		virtual void get_peer_info(peer_info& p) const;
+
 		// returns the torrent this connection is a part of
 		// may be zero if the connection is an incoming connection
 		// and it hasn't received enough information to determine
@@ -170,22 +213,20 @@ namespace libtorrent
 		// is called once every second by the main loop
 		void second_tick(float tick_interval);
 
-		boost::shared_ptr<stream_socket> get_socket() const { return m_socket; }
+		boost::shared_ptr<socket_type> get_socket() const { return m_socket; }
 		tcp::endpoint const& remote() const { return m_remote; }
 
 		std::vector<bool> const& get_bitfield() const;
 
+		void timed_out();
 		// this will cause this peer_connection to be disconnected.
-		// what it does is that it puts a reference to it in
-		// m_ses.m_disconnect_peer list, which will be scanned in the
-		// mainloop to disconnect peers.
 		void disconnect();
 		bool is_disconnecting() const { return m_disconnecting; }
 
 		// this is called when the connection attempt has succeeded
 		// and the peer_connection is supposed to set m_connecting
 		// to false, and stop monitor writability
-		void on_connection_complete(asio::error const& e);
+		void on_connection_complete(asio::error_code const& e);
 
 		// returns true if this connection is still waiting to
 		// finish the connection attempt
@@ -200,7 +241,7 @@ namespace libtorrent
 		// initiate the tcp connection. This may be postponed until
 		// the library isn't using up the limitation of half-open
 		// tcp connections.	
-		void connect();
+		void connect(int ticket);
 		
 		// This is called for every peer right after the upload
 		// bandwidth has been distributed among them
@@ -212,9 +253,8 @@ namespace libtorrent
 		void add_free_upload(size_type free_upload);
 
 		// trust management.
-		void received_valid_data();
-		void received_invalid_data();
-		int trust_points() const;
+		void received_valid_data(int index);
+		void received_invalid_data(int index);
 
 		size_type share_diff() const;
 
@@ -222,6 +262,10 @@ namespace libtorrent
 		// if it was an incoming connection, it is remote
 		bool is_local() const { return m_active; }
 
+		bool on_local_network() const;
+		bool ignore_bandwidth_limits() const
+		{ return m_ignore_bandwidth_limits; }
+
 		void set_failed() { m_failed = true; }
 		bool failed() const { return m_failed; }
 
@@ -261,20 +305,31 @@ namespace libtorrent
 
 		// adds a block to the request queue
 		void add_request(piece_block const& b);
+		// removes a block from the request queue or download queue
+		// sends a cancel message if appropriate
+		// refills the request queue, and possibly ignoring pieces requested
+		// by peers in the ignore list (to avoid recursion)
 		void cancel_request(piece_block const& b);
 		void send_block_requests();
 
-		// how much bandwidth we're using, how much we want,
-		// and how much we are allowed to use.
-		resource_request m_ul_bandwidth_quota;
-		resource_request m_dl_bandwidth_quota;
+		int max_assignable_bandwidth(int channel) const
+		{ return m_bandwidth_limit[channel].max_assignable(); }
+		
+		int bandwidth_throttle(int channel) const
+		{ return m_bandwidth_limit[channel].throttle(); }
+
+		void assign_bandwidth(int channel, int amount);
+		void expire_bandwidth(int channel, int amount);
 
 #ifndef NDEBUG
 		void check_invariant() const;
-		boost::posix_time::ptime m_last_choke;
+		ptime m_last_choke;
 #endif
 
-		virtual void get_peer_info(peer_info& p) const = 0;
+
+		// is true until we can be sure that the other end
+		// speaks our protocol (be it bittorrent or http).
+		virtual bool in_handshake() const = 0;
 
 		// returns the block currently being
 		// downloaded. And the progress of that
@@ -290,8 +345,24 @@ namespace libtorrent
 			return boost::optional<piece_block_progress>();
 		}
 
+		void send_buffer(char const* begin, char const* end);
+		buffer::interval allocate_send_buffer(int size);
+		void setup_send();
+
+#ifndef TORRENT_DISABLE_RESOLVE_COUNTRIES	
+		void set_country(char const* c)
+		{
+			assert(strlen(c) == 2);
+			m_country[0] = c[0];
+			m_country[1] = c[1];
+		}
+		bool has_country() const { return m_country[0] != 0; }
+#endif
+
 	protected:
 
+		virtual void get_specific_peer_info(peer_info& p) const = 0;
+
 		virtual void write_choke() = 0;
 		virtual void write_unchoke() = 0;
 		virtual void write_interested() = 0;
@@ -300,18 +371,16 @@ namespace libtorrent
 		virtual void write_cancel(peer_request const& r) = 0;
 		virtual void write_have(int index) = 0;
 		virtual void write_keepalive() = 0;
-		virtual void write_piece(peer_request const& r) = 0;
+		virtual void write_piece(peer_request const& r, char const* buffer) = 0;
 		
 		virtual void on_connected() = 0;
 		virtual void on_tick() {}
 	
-		virtual void on_receive(asio::error const& error
+		virtual void on_receive(asio::error_code const& error
 			, std::size_t bytes_transferred) = 0;
-		virtual void on_sent(asio::error const& error
+		virtual void on_sent(asio::error_code const& error
 			, std::size_t bytes_transferred) = 0;
 
-		void send_buffer(char const* begin, char const* end);
-		buffer::interval allocate_send_buffer(int size);
 		int send_buffer_size() const
 		{
 			return (int)m_send_buffer[0].size()
@@ -319,8 +388,18 @@ namespace libtorrent
 				- m_write_pos;
 		}
 
+#ifndef TORRENT_DISABLE_ENCRYPTION
+		buffer::interval wr_recv_buffer()
+		{
+			if (m_recv_buffer.empty()) return buffer::interval(0,0);
+			return buffer::interval(&m_recv_buffer[0]
+				, &m_recv_buffer[0] + m_recv_pos);
+		}
+#endif
+		
 		buffer::const_interval receive_buffer() const
 		{
+			if (m_recv_buffer.empty()) return buffer::const_interval(0,0);
 			return buffer::const_interval(&m_recv_buffer[0]
 				, &m_recv_buffer[0] + m_recv_pos);
 		}
@@ -332,17 +411,19 @@ namespace libtorrent
 
 		bool packet_finished() const
 		{
-			assert(m_recv_pos <= m_packet_size);
-			return m_packet_size == m_recv_pos;
+			return m_packet_size <= m_recv_pos;
 		}
 
-		void setup_send();
 		void setup_receive();
 
 		void attach_to_torrent(sha1_hash const& ih);
 
 		bool verify_piece(peer_request const& p) const;
 
+		// the bandwidth channels, upload and download
+		// keeps track of the current quotas
+		bandwidth_limit m_bandwidth_limit[num_channels];
+
 		// statistics about upload and download speeds
 		// and total amount of uploads and downloads for
 		// this peer
@@ -357,9 +438,9 @@ namespace libtorrent
 		
 		// called from the main loop when this connection has any
 		// work to do.
-		void on_send_data(asio::error const& error
+		void on_send_data(asio::error_code const& error
 			, std::size_t bytes_transferred);
-		void on_receive_data(asio::error const& error
+		void on_receive_data(asio::error_code const& error
 			, std::size_t bytes_transferred);
 
 		// this is the limit on the number of outstanding requests
@@ -372,16 +453,35 @@ namespace libtorrent
 
 		void set_timeout(int s) { m_timeout = s; }
 
+#ifndef TORRENT_DISABLE_EXTENSIONS
+		typedef std::list<boost::shared_ptr<peer_plugin> > extension_list_t;
+		extension_list_t m_extensions;
+#endif
+
+#ifndef TORRENT_DISABLE_RESOLVE_COUNTRIES	
+		// in case the session settings is set
+		// to resolve countries, this is set to
+		// the two character country code this
+		// peer resides in.
+		char m_country[2];
+#endif
+
 	private:
 
 		void fill_send_buffer();
+		void on_disk_read_complete(int ret, disk_io_job const& j, peer_request r);
+		void on_disk_write_complete(int ret, disk_io_job const& j
+			, peer_request r, boost::shared_ptr<torrent> t);
 
 		// the timeout in seconds
 		int m_timeout;
 
 		// the time when we last got a part of a
 		// piece packet from this peer
-		boost::posix_time::ptime m_last_piece;
+		ptime m_last_piece;
+		// the time we sent a request to
+		// this peer the last time
+		ptime m_last_request;
 
 		int m_packet_size;
 		int m_recv_pos;
@@ -401,6 +501,11 @@ namespace libtorrent
 		// (m_current_send_buffer + 1) % 2 is the
 		// buffer we're currently waiting for.
 		int m_current_send_buffer;
+
+		// the number of bytes we are currently reading
+		// from disk, that will be added to the send
+		// buffer as soon as they complete
+		int m_reading_bytes;
 		
 		// if the sending buffer doesn't finish in one send
 		// operation, this is the position within that buffer
@@ -408,12 +513,15 @@ namespace libtorrent
 		int m_write_pos;
 
 		// timeouts
-		boost::posix_time::ptime m_last_receive;
-		boost::posix_time::ptime m_last_sent;
+		ptime m_last_receive;
+		ptime m_last_sent;
 
-		boost::shared_ptr<stream_socket> m_socket;
+		boost::shared_ptr<socket_type> m_socket;
+		// this is the peer we're actually talking to
+		// it may not necessarily be the peer we're
+		// connected to, in case we use a proxy
 		tcp::endpoint m_remote;
-
+		
 		// this is the torrent this connection is
 		// associated with. If the connection is an
 		// incoming conncetion, this is set to zero
@@ -448,6 +556,11 @@ namespace libtorrent
 		// this peer
 		bool m_failed;
 
+		// if this is set to true, the peer will not
+		// request bandwidth from the limiter, but instead
+		// just send and receive as much as possible.
+		bool m_ignore_bandwidth_limits;
+
 		// the pieces the other end have
 		std::vector<bool> m_have_piece;
 
@@ -481,14 +594,6 @@ namespace libtorrent
 		// that we give the free upload, to keep the balance.
 		size_type m_free_upload;
 
-		// for every valid piece we receive where this
-		// peer was one of the participants, we increase
-		// this value. For every invalid piece we receive
-		// where this peer was a participant, we decrease
-		// this value. If it sinks below a threshold, its
-		// considered a bad peer and will be banned.
-		int m_trust_points;
-
 		// if this is true, this peer is assumed to handle all piece
 		// requests in fifo order. All skipped blocks are re-requested
 		// immediately instead of having a looser requirement
@@ -511,11 +616,11 @@ namespace libtorrent
 
 		// the time when this peer sent us a not_interested message
 		// the last time.
-		boost::posix_time::ptime m_became_uninterested;
+		ptime m_became_uninterested;
 
 		// the time when we sent a not_interested message to
 		// this peer the last time.
-		boost::posix_time::ptime m_became_uninteresting;
+		ptime m_became_uninteresting;
 
 		// this is true until this socket has become
 		// writable for the first time (i.e. the
@@ -533,19 +638,67 @@ namespace libtorrent
 		bool m_queued;
 
 		// these are true when there's a asynchronous write
-		// or read operation running.
+		// or read operation in progress. Or an asyncronous bandwidth
+		// request is in progress.
 		bool m_writing;
-		// this is the number of bytes sent to the socket last
-		// time it was invoked. This is compared against the
-		// bytes_transferred in the callback function that tells
-		// how much actually was sent. Then the quota can be
-		// corrected according to the actual number of bytes sent
-		int m_last_write_size;
 		bool m_reading;
-		int m_last_read_size;		
-		// reference counter for intrusive_ptr
-		mutable boost::detail::atomic_count m_refs;
 
+		// if set to true, this peer will always prefer
+		// to request entire pieces, rather than blocks.
+		// if it is false, the download rate limit setting
+		// will be used to determine if whole pieces
+		// are preferred.
+		bool m_prefer_whole_pieces;
+		
+		// if this is true, the blocks picked by the piece
+		// picker will be merged before passed to the
+		// request function. i.e. subsequent blocks are
+		// merged into larger blocks. This is used by
+		// the http-downloader, to request whole pieces
+		// at a time.
+		bool m_request_large_blocks;
+		
+		// if this is true, other (prioritized) peers will
+		// skip ahead of it in the queue for bandwidth. The
+		// effect is that non prioritized peers will only use
+		// the left-over bandwidth (suitable for web seeds).
+		bool m_non_prioritized;
+
+		int m_upload_limit;
+		int m_download_limit;
+
+		// this peer's peer info struct. This may
+		// be 0, in case the connection is incoming
+		// and hasn't been added to a torrent yet.
+		policy::peer* m_peer_info;
+
+		// this is a measurement of how fast the peer
+		// it allows some variance without changing
+		// back and forth between states
+		peer_speed_t m_speed;
+
+		// the ticket id from the connection queue.
+		// This is used to identify the connection
+		// so that it can be removed from the queue
+		// once the connection completes
+		int m_connection_ticket;
+		
+		// bytes downloaded since last second
+		// timer timeout; used for determining 
+		// approx download rate
+		int m_remote_bytes_dled;
+
+		// approximate peer download rate
+		int m_remote_dl_rate;
+
+		// a timestamp when the remote download rate
+		// was last updated
+		ptime m_remote_dl_update;
+
+		// the number of bytes send to the disk-io
+		// thread that hasn't yet been completely written.
+		int m_outstanding_writing_bytes;
+		
 #ifndef NDEBUG
 	public:
 		bool m_in_constructor;
