@@ -30,13 +30,13 @@ POSSIBILITY OF SUCH DAMAGE.
 
 */
 
+#include "libtorrent/pch.hpp"
+
 #include <fstream>
 #include <set>
 #include <numeric>
 #include <stdexcept>
 #include <boost/bind.hpp>
-#include <boost/date_time/posix_time/posix_time_types.hpp>
-#include <boost/date_time/posix_time/ptime.hpp>
 #include <boost/ref.hpp>
 #include <boost/optional.hpp>
 #include <boost/lexical_cast.hpp>
@@ -52,14 +52,6 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/io.hpp"
 #include "libtorrent/version.hpp"
 
-using boost::posix_time::ptime;
-using boost::posix_time::time_duration;
-using boost::posix_time::second_clock;
-using boost::posix_time::microsec_clock;
-using boost::posix_time::seconds;
-using boost::posix_time::minutes;
-using boost::posix_time::hours;
-using boost::posix_time::milliseconds;
 using boost::ref;
 using boost::lexical_cast;
 using libtorrent::dht::node_impl;
@@ -70,6 +62,11 @@ using libtorrent::dht::packet_iterator;
 namespace messages = libtorrent::dht::messages;
 using namespace libtorrent::detail;
 
+enum
+{
+	key_refresh = 5 // generate a new write token key every 5 minutes
+};
+
 using asio::ip::udp;
 typedef asio::ip::address_v4 address;
 
@@ -126,26 +123,43 @@ namespace
 
 namespace libtorrent { namespace dht
 {
+
+	void intrusive_ptr_add_ref(dht_tracker const* c)
+	{
+		assert(c != 0);
+		assert(c->m_refs >= 0);
+		++c->m_refs;
+	}
+
+	void intrusive_ptr_release(dht_tracker const* c)
+	{
+		assert(c != 0);
+		assert(c->m_refs > 0);
+		if (--c->m_refs == 0)
+			delete c;
+	}
+
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
 	TORRENT_DEFINE_LOG(dht_tracker)
 #endif
 
 	// class that puts the networking and the kademlia node in a single
 	// unit and connecting them together.
-	dht_tracker::dht_tracker(asio::io_service& d, dht_settings const& settings
+	dht_tracker::dht_tracker(asio::io_service& ios, dht_settings const& settings
 		, asio::ip::address listen_interface, entry const& bootstrap)
-		: m_demuxer(d)
-		, m_socket(m_demuxer, udp::endpoint(listen_interface, settings.service_port))
+		: m_strand(ios)
+		, m_socket(ios, udp::endpoint(listen_interface, settings.service_port))
 		, m_dht(bind(&dht_tracker::send_packet, this, _1), settings
 			, read_id(bootstrap))
 		, m_buffer(0)
-		, m_last_refresh(second_clock::universal_time() - hours(1))
-		, m_timer(m_demuxer)
-		, m_connection_timer(m_demuxer)
-		, m_refresh_timer(m_demuxer)
+		, m_last_new_key(time_now() - minutes(key_refresh))
+		, m_timer(ios)
+		, m_connection_timer(ios)
+		, m_refresh_timer(ios)
 		, m_settings(settings)
 		, m_refresh_bucket(160)
-		, m_host_resolver(d)
+		, m_host_resolver(ios)
+		, m_refs(0)
 	{
 		using boost::bind;
 
@@ -164,6 +178,7 @@ namespace libtorrent { namespace dht
 		m_lt_message_input = 0;
 		m_mp_message_input = 0;
 		m_gr_message_input = 0;
+		m_mo_message_input = 0;
 		m_total_in_bytes = 0;
 		m_total_out_bytes = 0;
 		m_queries_out_bytes = 0;
@@ -187,47 +202,62 @@ namespace libtorrent { namespace dht
 			} catch (std::exception&) {}
 		}
 
-		m_dht.bootstrap(initial_nodes, bind(&dht_tracker::on_bootstrap, this));
-
 		m_socket.async_receive_from(asio::buffer(&m_in_buf[m_buffer][0]
 			, m_in_buf[m_buffer].size()), m_remote_endpoint[m_buffer]
-			, bind(&dht_tracker::on_receive, this, _1, _2));
+			, m_strand.wrap(bind(&dht_tracker::on_receive, self(), _1, _2)));
 		m_timer.expires_from_now(seconds(1));
-		m_timer.async_wait(bind(&dht_tracker::tick, this, _1));
+		m_timer.async_wait(m_strand.wrap(bind(&dht_tracker::tick, self(), _1)));
 
 		m_connection_timer.expires_from_now(seconds(10));
-		m_connection_timer.async_wait(bind(&dht_tracker::connection_timeout, this, _1));
+		m_connection_timer.async_wait(m_strand.wrap(
+			bind(&dht_tracker::connection_timeout, self(), _1)));
 
-		m_refresh_timer.expires_from_now(minutes(15));
-		m_refresh_timer.async_wait(bind(&dht_tracker::refresh_timeout, this, _1));
+		m_refresh_timer.expires_from_now(seconds(5));
+		m_refresh_timer.async_wait(m_strand.wrap(bind(&dht_tracker::refresh_timeout, self(), _1)));
+
+		m_dht.bootstrap(initial_nodes, bind(&dht_tracker::on_bootstrap, self()));
+	}
+
+	void dht_tracker::stop()
+	{
+		m_timer.cancel();
+		m_connection_timer.cancel();
+		m_refresh_timer.cancel();
+		m_socket.close();
 	}
 
 	void dht_tracker::dht_status(session_status& s)
 	{
-		boost::tie(s.m_dht_nodes, s.m_dht_node_cache) = m_dht.size();
-		s.m_dht_torrents = m_dht.data_size();
+		boost::tie(s.dht_nodes, s.dht_node_cache) = m_dht.size();
+		s.dht_torrents = m_dht.data_size();
+		s.dht_global_nodes = m_dht.num_global_nodes();
 	}
 
-	void dht_tracker::connection_timeout(asio::error const& e)
+	void dht_tracker::connection_timeout(asio::error_code const& e)
 		try
 	{
 		if (e) return;
 		time_duration d = m_dht.connection_timeout();
 		m_connection_timer.expires_from_now(d);
-		m_connection_timer.async_wait(bind(&dht_tracker::connection_timeout, this, _1));
+		m_connection_timer.async_wait(m_strand.wrap(bind(&dht_tracker::connection_timeout, self(), _1)));
 	}
-	catch (std::exception&)
+	catch (std::exception& exc)
 	{
+#ifndef NDEBUG
+		std::cerr << "exception-type: " << typeid(exc).name() << std::endl;
+		std::cerr << "what: " << exc.what() << std::endl;
 		assert(false);
+#endif
 	};
 
-	void dht_tracker::refresh_timeout(asio::error const& e)
+	void dht_tracker::refresh_timeout(asio::error_code const& e)
 		try
 	{
 		if (e) return;
 		time_duration d = m_dht.refresh_timeout();
 		m_refresh_timer.expires_from_now(d);
-		m_refresh_timer.async_wait(bind(&dht_tracker::refresh_timeout, this, _1));
+		m_refresh_timer.async_wait(m_strand.wrap(
+			bind(&dht_tracker::refresh_timeout, self(), _1)));
 	}
 	catch (std::exception&)
 	{
@@ -237,18 +267,27 @@ namespace libtorrent { namespace dht
 	void dht_tracker::rebind(asio::ip::address listen_interface, int listen_port)
 	{
 		m_socket.close();
-		m_socket.open(asio::ip::udp::v4());
-		m_socket.bind(udp::endpoint(listen_interface, listen_port));
+		udp::endpoint ep(listen_interface, listen_port);
+		m_socket.open(ep.protocol());
+		m_socket.bind(ep);
 	}
 
-	void dht_tracker::tick(asio::error const& e)
+	void dht_tracker::tick(asio::error_code const& e)
 		try
 	{
 		if (e) return;
 		m_timer.expires_from_now(minutes(tick_period));
-		m_timer.async_wait(bind(&dht_tracker::tick, this, _1));
+		m_timer.async_wait(m_strand.wrap(bind(&dht_tracker::tick, this, _1)));
 
-		m_dht.new_write_key();
+		ptime now = time_now();
+		if (now - m_last_new_key > minutes(key_refresh))
+		{
+			m_last_new_key = now;
+			m_dht.new_write_key();
+#ifdef TORRENT_DHT_VERBOSE_LOGGING
+			TORRENT_LOG(dht_tracker) << time_now_string() << " new write key";
+#endif
+		}
 		
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
 		static bool first = true;
@@ -271,9 +310,7 @@ namespace libtorrent { namespace dht
 		if (first)
 		{
 			first = false;
-			using boost::posix_time::to_simple_string;
-			pc << "\n\n *****   starting log at " << to_simple_string(
-				second_clock::universal_time()) << "   *****\n\n"
+			pc << "\n\n *****   starting log at " << time_now_string() << "   *****\n\n"
 				<< "minute:active nodes:passive nodes"
 				":ping replies sent:ping queries recvd:ping"
 				":ping replies sent:ping queries recvd:ping"
@@ -313,6 +350,7 @@ namespace libtorrent { namespace dht
 			<< "\t" << (m_lt_message_input / float(tick_period))
 			<< "\t" << (m_mp_message_input / float(tick_period))
 			<< "\t" << (m_gr_message_input / float(tick_period))
+			<< "\t" << (m_mo_message_input / float(tick_period))
 			<< "\t" << (m_total_in_bytes / float(tick_period*60))
 			<< "\t" << (m_total_out_bytes / float(tick_period*60))
 			<< "\t" << (m_queries_out_bytes / float(tick_period*60))
@@ -346,7 +384,7 @@ namespace libtorrent { namespace dht
 
 	// translate bittorrent kademlia message into the generice kademlia message
 	// used by the library
-	void dht_tracker::on_receive(asio::error const& error, size_t bytes_transferred)
+	void dht_tracker::on_receive(asio::error_code const& error, size_t bytes_transferred)
 		try
 	{
 		if (error == asio::error::operation_aborted) return;
@@ -355,7 +393,7 @@ namespace libtorrent { namespace dht
 		m_buffer = (m_buffer + 1) & 1;
 		m_socket.async_receive_from(asio::buffer(&m_in_buf[m_buffer][0]
 			, m_in_buf[m_buffer].size()), m_remote_endpoint[m_buffer]
-			, bind(&dht_tracker::on_receive, this, _1, _2));
+			, m_strand.wrap(bind(&dht_tracker::on_receive, self(), _1, _2)));
 
 		if (error) return;
 
@@ -375,9 +413,8 @@ namespace libtorrent { namespace dht
 				, m_in_buf[current_buffer].end());
 
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
-			TORRENT_LOG(dht_tracker) << microsec_clock::universal_time()
-				<< " RECEIVED [" << m_remote_endpoint[current_buffer]
-				<< "]:";
+			TORRENT_LOG(dht_tracker) << time_now_string() << " RECEIVED ["
+				<< m_remote_endpoint[current_buffer] << "]:";
 #endif
 
 			libtorrent::dht::msg m;
@@ -412,9 +449,14 @@ namespace libtorrent { namespace dht
 					++m_gr_message_input;
 					TORRENT_LOG(dht_tracker) << "   client: GetRight";
 				}
+				else if (client.size() > 1 && std::equal(client.begin(), client.begin() + 2, "MO"))
+				{
+					++m_mo_message_input;
+					TORRENT_LOG(dht_tracker) << "   client: Mono Torrent";
+				}
 				else
 				{
-					TORRENT_LOG(dht_tracker) << "   client: generic";
+					TORRENT_LOG(dht_tracker) << "   client: " << client;
 				}
 			}
 			catch (std::exception&)
@@ -438,10 +480,24 @@ namespace libtorrent { namespace dht
 				if (id.size() != 20) throw std::runtime_error("invalid size of id");
 				std::copy(id.begin(), id.end(), m.id.begin());
 
-				if (entry const* n = r.find_key("values"))				
+				if (entry const* n = r.find_key("values"))
 				{
 					m.peers.clear();
-					read_endpoint_list<tcp::endpoint>(n, m.peers);
+					if (n->list().size() == 1)
+					{
+						// assume it's mainline format
+						std::string const& peers = n->list().front().string();
+						std::string::const_iterator i = peers.begin();
+						std::string::const_iterator end = peers.end();
+
+						while (std::distance(i, end) >= 6)
+							m.peers.push_back(read_v4_endpoint<tcp::endpoint>(i));
+					}
+					else
+					{
+						// assume it's uTorrent/libtorrent format
+						read_endpoint_list<tcp::endpoint>(n, m.peers);
+					}
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
 					TORRENT_LOG(dht_tracker) << "   peers: " << m.peers.size();
 #endif
@@ -572,9 +628,10 @@ namespace libtorrent { namespace dht
 				m.error_msg = list.back().string();
 				m.error_code = list.front().integer();
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
-				TORRENT_LOG(dht_tracker) << "   error: " << m.error_code << " "
+				TORRENT_LOG(dht_tracker) << "   incoming error: " << m.error_code << " "
 					<< m.error_msg;
 #endif
+				throw std::runtime_error("DHT error message");
 			}
 			else
 			{
@@ -593,14 +650,17 @@ namespace libtorrent { namespace dht
 			}
 			TORRENT_LOG(dht_tracker) << e;
 #endif
-
+			assert(m.message_id != messages::error);
 			m_dht.incoming(m);
 		}
 		catch (std::exception& e)
 		{
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
+			int current_buffer = (m_buffer + 1) & 1;
+			std::string msg(m_in_buf[current_buffer].begin()
+				, m_in_buf[current_buffer].begin() + bytes_transferred);
 			TORRENT_LOG(dht_tracker) << "invalid incoming packet: "
-				<< e.what();
+				<< e.what() << "\n" << msg << "\n";
 #endif
 		}
 	}
@@ -648,11 +708,11 @@ namespace libtorrent { namespace dht
 	void dht_tracker::add_node(std::pair<std::string, int> const& node)
 	{
 		udp::resolver::query q(node.first, lexical_cast<std::string>(node.second));
-		m_host_resolver.async_resolve(q, bind(&dht_tracker::on_name_lookup
-			, this, _1, _2));
+		m_host_resolver.async_resolve(q, m_strand.wrap(
+			bind(&dht_tracker::on_name_lookup, self(), _1, _2)));
 	}
 
-	void dht_tracker::on_name_lookup(asio::error const& e
+	void dht_tracker::on_name_lookup(asio::error_code const& e
 		, udp::resolver::iterator host) try
 	{
 		if (e || host == udp::resolver::iterator()) return;
@@ -666,11 +726,11 @@ namespace libtorrent { namespace dht
 	void dht_tracker::add_router_node(std::pair<std::string, int> const& node)
 	{
 		udp::resolver::query q(node.first, lexical_cast<std::string>(node.second));
-		m_host_resolver.async_resolve(q, bind(&dht_tracker::on_router_name_lookup
-			, this, _1, _2));
+		m_host_resolver.async_resolve(q, m_strand.wrap(
+			bind(&dht_tracker::on_router_name_lookup, self(), _1, _2)));
 	}
 
-	void dht_tracker::on_router_name_lookup(asio::error const& e
+	void dht_tracker::on_router_name_lookup(asio::error_code const& e
 		, udp::resolver::iterator host) try
 	{
 		if (e || host == udp::resolver::iterator()) return;
@@ -684,23 +744,66 @@ namespace libtorrent { namespace dht
 	void dht_tracker::on_bootstrap()
 	{}
 
+	namespace
+	{
+		void write_nodes_entry(entry& r, libtorrent::dht::msg const& m)
+		{
+			bool ipv6_nodes = false;
+			r["nodes"] = entry(entry::string_t);
+			entry& n = r["nodes"];
+			std::back_insert_iterator<std::string> out(n.string());
+			for (msg::nodes_t::const_iterator i = m.nodes.begin()
+				, end(m.nodes.end()); i != end; ++i)
+			{
+				if (!i->addr.address().is_v4())
+				{
+					ipv6_nodes = true;
+					continue;
+				}
+				std::copy(i->id.begin(), i->id.end(), out);
+				write_endpoint(i->addr, out);
+			}
+
+			if (ipv6_nodes)
+			{
+				r["nodes2"] = entry(entry::list_t);
+				entry& p = r["nodes2"];
+				std::string endpoint;
+				for (msg::nodes_t::const_iterator i = m.nodes.begin()
+					, end(m.nodes.end()); i != end; ++i)
+				{
+					if (!i->addr.address().is_v6()) continue;
+					endpoint.resize(18 + 20);
+					std::string::iterator out = endpoint.begin();
+					std::copy(i->id.begin(), i->id.end(), out);
+					out += 20;
+					write_endpoint(i->addr, out);
+					endpoint.resize(out - endpoint.begin());
+					p.list().push_back(entry(endpoint));
+				}
+			}
+#ifdef TORRENT_DHT_VERBOSE_LOGGING
+			TORRENT_LOG(dht_tracker) << "   nodes: " << m.nodes.size();
+#endif
+		}
+	}
+
 	void dht_tracker::send_packet(msg const& m)
+		try
 	{
 		using libtorrent::bencode;
 		using libtorrent::entry;
 		entry e(entry::dictionary_t);
+		assert(!m.transaction_id.empty() || m.message_id == messages::error);
 		e["t"] = m.transaction_id;
-		std::string version_str("LT  ");
-		std::string::iterator i = version_str.begin() + 2;
-		detail::write_uint8(LIBTORRENT_VERSION_MAJOR, i);
-		detail::write_uint8(LIBTORRENT_VERSION_MINOR, i);
-		e["v"] = version_str;
+		static char const version_str[] = {'L', 'T'
+			, LIBTORRENT_VERSION_MAJOR, LIBTORRENT_VERSION_MINOR};
+		e["v"] = std::string(version_str, version_str + 4);
 
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
-		TORRENT_LOG(dht_tracker) << microsec_clock::universal_time()
+		TORRENT_LOG(dht_tracker) << time_now_string()
 			<< " SENDING [" << m.addr << "]:";
 		TORRENT_LOG(dht_tracker) << "   transaction: " << m.transaction_id;
-//		e.print(std::cerr);
 #endif
 
 		if (m.message_id == messages::error)
@@ -708,12 +811,13 @@ namespace libtorrent { namespace dht
 			assert(m.reply);
 			e["y"] = "e";
 			entry error_list(entry::list_t);
+			assert(m.error_code > 200 && m.error_code <= 204);
 			error_list.list().push_back(entry(m.error_code));
 			error_list.list().push_back(entry(m.error_msg));
 			e["e"] = error_list;
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
-		TORRENT_LOG(dht_tracker) << "   error: " << m.error_code << " "
-			<< m.error_msg;
+			TORRENT_LOG(dht_tracker) << time_now_string()
+				<< "   outgoing error: " << m.error_code << " " << m.error_msg;
 #endif
 		}
 		else if (m.reply)
@@ -724,8 +828,8 @@ namespace libtorrent { namespace dht
 			r["id"] = std::string(m.id.begin(), m.id.end());
 
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
-			TORRENT_LOG(dht_tracker) << "   reply: "
-				<< messages::ids[m.message_id];
+			TORRENT_LOG(dht_tracker) << time_now_string()
+				<< "   reply: " << messages::ids[m.message_id];
 #endif
 
 			if (m.write_token.type() != entry::undefined_t)
@@ -737,71 +841,27 @@ namespace libtorrent { namespace dht
 					break;
 				case messages::find_node:
 				{
-					bool ipv6_nodes = false;
-					r["nodes"] = entry(entry::string_t);
-					entry& n = r["nodes"];
-					std::back_insert_iterator<std::string> out(n.string());
-					for (msg::nodes_t::const_iterator i = m.nodes.begin()
-						, end(m.nodes.end()); i != end; ++i)
-					{
-						if (!i->addr.address().is_v4())
-						{
-							ipv6_nodes = true;
-							continue;
-						}
-						std::copy(i->id.begin(), i->id.end(), out);
-						write_endpoint(i->addr, out);
-					}
-
-					if (ipv6_nodes)
-					{
-						r["nodes2"] = entry(entry::list_t);
-						entry& p = r["nodes2"];
-						std::string endpoint;
-						endpoint.resize(6);
-						for (msg::nodes_t::const_iterator i = m.nodes.begin()
-							, end(m.nodes.end()); i != end; ++i)
-						{
-							std::string::iterator out = endpoint.begin();
-							std::copy(i->id.begin(), i->id.end(), out);
-							write_endpoint(i->addr, out);
-							p.list().push_back(entry(endpoint));
-						}
-					}
-#ifdef TORRENT_DHT_VERBOSE_LOGGING
-					TORRENT_LOG(dht_tracker) << "   nodes: " << m.nodes.size();
-#endif
+					write_nodes_entry(r, m);
 					break;
 				}
 				case messages::get_peers:
 				{
 					if (m.peers.empty())
 					{
-						r["nodes"] = entry(entry::string_t);
-						entry& n = r["nodes"];
-						std::back_insert_iterator<std::string> out(n.string());
-						for (msg::nodes_t::const_iterator i = m.nodes.begin()
-							, end(m.nodes.end()); i != end; ++i)
-						{
-							if (!i->addr.address().is_v4()) continue;
-							std::copy(i->id.begin(), i->id.end(), out);
-							write_endpoint(i->addr, out);
-						}
-#ifdef TORRENT_DHT_VERBOSE_LOGGING
-						TORRENT_LOG(dht_tracker) << "   nodes: " << m.nodes.size();
-#endif
+						write_nodes_entry(r, m);
 					}
 					else
 					{
 						r["values"] = entry(entry::list_t);
 						entry& p = r["values"];
 						std::string endpoint;
-						endpoint.resize(6);
 						for (msg::peers_t::const_iterator i = m.peers.begin()
 							, end(m.peers.end()); i != end; ++i)
 						{
+							endpoint.resize(18);
 							std::string::iterator out = endpoint.begin();
 							write_endpoint(*i, out);
+							endpoint.resize(out - endpoint.begin());
 							p.list().push_back(entry(endpoint));
 						}
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
@@ -855,7 +915,7 @@ namespace libtorrent { namespace dht
 				}
 				case messages::announce_peer:
 					a["port"] = m_settings.service_port;
-					a["info_hash"] = boost::lexical_cast<std::string>(m.info_hash);
+					a["info_hash"] = std::string(m.info_hash.begin(), m.info_hash.end());
 					a["token"] = m.write_token;
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
 					TORRENT_LOG(dht_tracker) << "   port: "
@@ -870,8 +930,10 @@ namespace libtorrent { namespace dht
 
 		m_send_buf.clear();
 		bencode(std::back_inserter(m_send_buf), e);
+		asio::error_code ec;
 		m_socket.send_to(asio::buffer(&m_send_buf[0]
-			, (int)m_send_buf.size()), m.addr);
+			, (int)m_send_buf.size()), m.addr, 0, ec);
+		if (ec) return;
 
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
 		m_total_out_bytes += m_send_buf.size();
@@ -900,6 +962,13 @@ namespace libtorrent { namespace dht
 	
 		send_packet(pm);	
 	}
+	catch (std::exception&)
+	{
+		// m_send may fail with "no route to host"
+		// but it shouldn't throw since an error code
+		// is passed in instead
+		assert(false);
+	}
 
 }}
 
