@@ -30,6 +30,8 @@ POSSIBILITY OF SUCH DAMAGE.
 
 */
 
+#include "libtorrent/pch.hpp"
+
 #include <ctime>
 #include <iostream>
 #include <fstream>
@@ -43,7 +45,6 @@ POSSIBILITY OF SUCH DAMAGE.
 #endif
 
 #include <boost/lexical_cast.hpp>
-#include <boost/date_time/gregorian/gregorian_types.hpp>
 #include <boost/filesystem/path.hpp>
 #include <boost/next_prior.hpp>
 #include <boost/bind.hpp>
@@ -57,11 +58,16 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/hasher.hpp"
 #include "libtorrent/entry.hpp"
 
+namespace pt = boost::posix_time;
+namespace gr = boost::gregorian;
+
 using namespace libtorrent;
-using namespace boost::filesystem;
 
 namespace
 {
+	
+	namespace fs = boost::filesystem;
+
 	void convert_to_utf8(std::string& str, unsigned char chr)
 	{
 		str += 0xc0 | ((chr & 0xff) >> 6);
@@ -149,7 +155,7 @@ namespace
 		// encoded string
 		if (!valid_encoding)
 		{
-			target.orig_path.reset(new path(target.path));
+			target.orig_path.reset(new fs::path(target.path));
 			target.path = tmp_path;
 		}
 	}
@@ -199,8 +205,8 @@ namespace
 			offset += target.back().size;
 		}
 	}
-
-	void remove_dir(path& p)
+/*
+	void remove_dir(fs::path& p)
 	{
 		assert(p.begin() != p.end());
 		path tmp;
@@ -208,20 +214,22 @@ namespace
 			tmp /= *i;
 		p = tmp;
 	}
+*/
 }
 
 namespace libtorrent
 {
 
-	using namespace boost::gregorian;
-	using namespace boost::posix_time;
-
 	// standard constructor that parses a torrent file
 	torrent_info::torrent_info(const entry& torrent_file)
-		: m_creation_date(date(not_a_date_time))
+		: m_num_pieces(0)
+		, m_creation_date(pt::ptime(pt::not_a_date_time))
 		, m_multifile(false)
 		, m_private(false)
 		, m_extra_info(entry::dictionary_t)
+#ifndef NDEBUG
+		, m_half_metadata(false)
+#endif
 	{
 		try
 		{
@@ -238,30 +246,63 @@ namespace libtorrent
 	// just the necessary to use it with piece manager
 	// used for torrents with no metadata
 	torrent_info::torrent_info(sha1_hash const& info_hash)
-		: m_piece_length(256 * 1024)
+		: m_piece_length(0)
 		, m_total_size(0)
+		, m_num_pieces(0)
 		, m_info_hash(info_hash)
 		, m_name()
-		, m_creation_date(second_clock::universal_time())
+		, m_creation_date(pt::second_clock::universal_time())
 		, m_multifile(false)
+		, m_private(false)
 		, m_extra_info(entry::dictionary_t)
+#ifndef NDEBUG
+		, m_half_metadata(false)
+#endif
 	{
 	}
 
 	torrent_info::torrent_info()
-		: m_piece_length(256 * 1024)
+		: m_piece_length(0)
 		, m_total_size(0)
+		, m_num_pieces(0)
 		, m_info_hash(0)
 		, m_name()
-		, m_creation_date(second_clock::universal_time())
+		, m_creation_date(pt::second_clock::universal_time())
 		, m_multifile(false)
+		, m_private(false)
 		, m_extra_info(entry::dictionary_t)
+#ifndef NDEBUG
+		, m_half_metadata(false)
+#endif
 	{
 	}
 
 	torrent_info::~torrent_info()
 	{}
 
+	void torrent_info::swap(torrent_info& ti)
+	{
+		using std::swap;
+		m_urls.swap(ti.m_urls);
+		m_url_seeds.swap(ti.m_url_seeds);
+		swap(m_piece_length, ti.m_piece_length);
+		m_piece_hash.swap(ti.m_piece_hash);
+		m_files.swap(ti.m_files);
+		m_nodes.swap(ti.m_nodes);
+		swap(m_num_pieces, ti.m_num_pieces);
+		swap(m_info_hash, ti.m_info_hash);
+		m_name.swap(ti.m_name);
+		swap(m_creation_date, ti.m_creation_date);
+		m_comment.swap(ti.m_comment);
+		m_created_by.swap(ti.m_created_by);
+		swap(m_multifile, ti.m_multifile);
+		swap(m_private, ti.m_private);
+		m_extra_info.swap(ti.m_extra_info);
+#ifndef NDEBUG
+		swap(m_half_metadata, ti.m_half_metadata);
+#endif
+	}
+
 	void torrent_info::set_piece_size(int size)
 	{
 		// make sure the size is an even power of 2
@@ -275,15 +316,15 @@ namespace libtorrent
 			}
 		}
 #endif
+		assert(!m_half_metadata);
 		m_piece_length = size;
 
-
-		int num_pieces = static_cast<int>(
+		m_num_pieces = static_cast<int>(
 			(m_total_size + m_piece_length - 1) / m_piece_length);
 		int old_num_pieces = static_cast<int>(m_piece_hash.size());
 
-		m_piece_hash.resize(num_pieces);
-		for (int i = old_num_pieces; i < num_pieces; ++i)
+		m_piece_hash.resize(m_num_pieces);
+		for (int i = old_num_pieces; i < m_num_pieces; ++i)
 		{
 			m_piece_hash[i].clear();
 		}
@@ -308,7 +349,7 @@ namespace libtorrent
 		else
 		{ m_name = info["name"].string(); }
 		
-		path tmp = m_name;
+		fs::path tmp = m_name;
 		if (tmp.is_complete()) throw std::runtime_error("torrent contains "
 			"a file with an absolute path: '" + m_name + "'");
 		if (tmp.has_branch_path()) throw std::runtime_error(
@@ -341,14 +382,14 @@ namespace libtorrent
 		// we want this division to round upwards, that's why we have the
 		// extra addition
 
-		int num_pieces = static_cast<int>((m_total_size + m_piece_length - 1) / m_piece_length);
-		m_piece_hash.resize(num_pieces);
+		m_num_pieces = static_cast<int>((m_total_size + m_piece_length - 1) / m_piece_length);
+		m_piece_hash.resize(m_num_pieces);
 		const std::string& hash_string = info["pieces"].string();
 
-		if ((int)hash_string.length() != num_pieces * 20)
+		if ((int)hash_string.length() != m_num_pieces * 20)
 			throw invalid_torrent_file();
 
-		for (int i = 0; i < num_pieces; ++i)
+		for (int i = 0; i < m_num_pieces; ++i)
 			std::copy(
 				hash_string.begin() + i*20
 				, hash_string.begin() + (i+1)*20
@@ -451,8 +492,8 @@ namespace libtorrent
 		// extract creation date
 		try
 		{
-			m_creation_date = ptime(date(1970, Jan, 1))
-				+ seconds(long(torrent_file["creation date"].integer()));
+			m_creation_date = pt::ptime(gr::date(1970, gr::Jan, 1))
+				+ pt::seconds(long(torrent_file["creation date"].integer()));
 		}
 		catch (type_error) {}
 
@@ -490,14 +531,14 @@ namespace libtorrent
 		parse_info_section(torrent_file["info"]);
 	}
 
-	boost::optional<ptime>
+	boost::optional<pt::ptime>
 	torrent_info::creation_date() const
 	{
-		if (m_creation_date != ptime(date(not_a_date_time)))
+		if (m_creation_date != pt::ptime(gr::date(pt::not_a_date_time)))
 		{
-			return boost::optional<ptime>(m_creation_date);
+			return boost::optional<pt::ptime>(m_creation_date);
 		}
-		return boost::optional<ptime>();
+		return boost::optional<pt::ptime>();
 	}
 
 	void torrent_info::add_tracker(std::string const& url, int tier)
@@ -511,9 +552,9 @@ namespace libtorrent
 			, bind(&announce_entry::tier, _1), bind(&announce_entry::tier, _2)));
 	}
 
-	void torrent_info::add_file(boost::filesystem::path file, size_type size)
+	void torrent_info::add_file(fs::path file, size_type size)
 	{
-		assert(file.begin() != file.end());
+//		assert(file.begin() != file.end());
 
 		if (!file.has_branch_path())
 		{
@@ -538,21 +579,23 @@ namespace libtorrent
 		file_entry e;
 		e.path = file;
 		e.size = size;
+		e.offset = m_files.empty() ? 0 : m_files.back().offset
+			+ m_files.back().size;
 		m_files.push_back(e);
 
 		m_total_size += size;
+		
+		if (m_piece_length == 0)
+			m_piece_length = 256 * 1024;
 
-		int num_pieces = static_cast<int>(
+		m_num_pieces = static_cast<int>(
 			(m_total_size + m_piece_length - 1) / m_piece_length);
 		int old_num_pieces = static_cast<int>(m_piece_hash.size());
 
-		m_piece_hash.resize(num_pieces);
-		for (std::vector<sha1_hash>::iterator i = m_piece_hash.begin() + old_num_pieces;
-			i != m_piece_hash.end(); ++i)
-		{
-			i->clear();
-		}
-
+		m_piece_hash.resize(m_num_pieces);
+		if (m_num_pieces > old_num_pieces)
+			std::for_each(m_piece_hash.begin() + old_num_pieces
+				, m_piece_hash.end(), boost::bind(&sha1_hash::clear, _1));
 	}
 
 	void torrent_info::add_url_seed(std::string const& url)
@@ -572,8 +615,6 @@ namespace libtorrent
 
 	entry torrent_info::create_info_metadata() const
 	{
-		namespace fs = boost::filesystem;
-
 		// you have to add files to the torrent first
 		assert(!m_files.empty());
 	
@@ -591,16 +632,14 @@ namespace libtorrent
 			if (!info.find_key("files"))
 			{
 				entry& files = info["files"];
-				files = entry(entry::list_t);
 
 				for (std::vector<file_entry>::const_iterator i = m_files.begin();
 					i != m_files.end(); ++i)
 				{
-					files.list().push_back(entry(entry::dictionary_t));
+					files.list().push_back(entry());
 					entry& file_e = files.list().back();
 					file_e["length"] = i->size;
 					entry& path_e = file_e["path"];
-					path_e = entry(entry::list_t);
 
 					fs::path const* file_path;
 					if (i->orig_path) file_path = &(*i->orig_path);
@@ -619,7 +658,6 @@ namespace libtorrent
 
 		info["piece length"] = piece_length();
 		entry& pieces = info["pieces"];
-		pieces = entry(entry::string_t);
 
 		std::string& p = pieces.string();
 
@@ -636,13 +674,6 @@ namespace libtorrent
 	{
 		assert(m_piece_length > 0);
 
-		using namespace boost::gregorian;
-		using namespace boost::posix_time;
-
-		namespace fs = boost::filesystem;
-
-		entry dict(entry::dictionary_t);
-
 		if ((m_urls.empty() && m_nodes.empty()) || m_files.empty())
 		{
 			// TODO: throw something here
@@ -650,6 +681,8 @@ namespace libtorrent
 			return entry();
 		}
 
+		entry dict;
+
 		if (m_private) dict["private"] = 1;
 
 		if (!m_urls.empty())
@@ -658,7 +691,6 @@ namespace libtorrent
 		if (!m_nodes.empty())
 		{
 			entry& nodes = dict["nodes"];
-			nodes = entry(entry::list_t);
 			entry::list_type& nodes_list = nodes.list();
 			for (nodes_t::const_iterator i = m_nodes.begin()
 				, end(m_nodes.end()); i != end; ++i)
@@ -694,7 +726,7 @@ namespace libtorrent
 			dict["comment"] = m_comment;
 
 		dict["creation date"] =
-			(m_creation_date - ptime(date(1970, Jan, 1))).total_seconds();
+			(m_creation_date - pt::ptime(gr::date(1970, gr::Jan, 1))).total_seconds();
 
 		if (!m_created_by.empty())
 			dict["created by"] = m_created_by;
@@ -708,7 +740,6 @@ namespace libtorrent
 			else
 			{
 				entry& list = dict["url-list"];
-				list = entry(entry::list_t);
 				for (std::vector<std::string>::const_iterator i
 					= m_url_seeds.begin(); i != m_url_seeds.end(); ++i)
 				{
@@ -739,6 +770,18 @@ namespace libtorrent
 		assert(false);
 	}
 
+	void torrent_info::seed_free()
+	{
+		std::vector<std::string>().swap(m_url_seeds);
+		nodes_t().swap(m_nodes);
+		std::vector<sha1_hash>().swap(m_piece_hash);
+#ifndef NDEBUG
+		m_half_metadata = true;
+#endif
+	}
+
+// ------- start deprecation -------
+
 	void torrent_info::print(std::ostream& os) const
 	{
 		os << "trackers:\n";
@@ -749,8 +792,8 @@ namespace libtorrent
 		}
 		if (!m_comment.empty())
 			os << "comment: " << m_comment << "\n";
-		if (m_creation_date != ptime(date(not_a_date_time)))
-			os << "creation date: " << to_simple_string(m_creation_date) << "\n";
+//		if (m_creation_date != pt::ptime(gr::date(pt::not_a_date_time)))
+//			os << "creation date: " << to_simple_string(m_creation_date) << "\n";
 		os << "private: " << (m_private?"yes":"no") << "\n";
 		os << "number of pieces: " << num_pieces() << "\n";
 		os << "piece length: " << piece_length() << "\n";
@@ -759,6 +802,8 @@ namespace libtorrent
 			os << "  " << std::setw(11) << i->size << "  " << i->path.string() << "\n";
 	}
 
+// ------- end deprecation -------
+
 	size_type torrent_info::piece_size(int index) const
 	{
 		assert(index >= 0 && index < num_pieces());
